/**
 * An object that implements basic file system functionalities.
 */
export interface FileSystem {
  /**
   * Asynchronously tests whether or not the given path exists by checking with the file system.
   * @param path A path to a file or directory.
   */
  exists(path: string): Promise<boolean>

  /**
   * Asynchronously reads the entire contents of a file.
   * @param path A path to a file.
   */
  readFile(path: string): Promise<Uint8Array>

  /**
   * Asynchronously reads the entire contents of a file.
   * @param path A path to a file.
   */
  readTextFile(path: string): Promise<string>

  /**
   * Asynchronously writes data to a file, replacing the file if it already exists.
   * @param path A path to a file.
   * @param data The data to write.
   */
  writeTextFile(path: string, data: string): Promise<void>

  /**
   * Reads the directory given by path and returns an async iterable of `DirEntry`.
   */
  readDir(path: string): AsyncIterable<DirEntry>
}

/**
 * Information about a directory entry
 */
export interface DirEntry {
  /** The file name of the entry. Does not include the full path. */
  name: string
  /** True if this is info for a regular file. */
  isFile: boolean
  /** True if this is info for a directory. */
  isDirectory: boolean
}

export interface Adapter {
  /**
   * Underlying `fs` module. You can replace it with a wrapper of
   * `memfs` or any object that implements `IFileSystem`.
   */
  fs: FileSystem

  /**
   * `SubtleCrypto` implementation. On Node.js this is
   * `require("crypto").webcrypto.subtle` and in the browser this is
   * `window.crypto.subtle`.
   */
  subtle: SubtleCrypto
}
