import type { Adapter, DirEntry, FileSystem } from "./index"

function normalize(path: string) {
  return path.replace(/^\//, "")
}

function splitPath(path: string) {
  const segments = normalize(path).split("/")
  const filename = segments.pop()!
  return [segments, filename] as const
}

class FS implements FileSystem {
  constructor(private handle: FileSystemDirectoryHandle) {}

  private async getDirectoryHandle(segments: string[]) {
    if (!segments.length || (segments.length === 1 && !segments[0])) {
      return this.handle
    }
    const [first, ...pathSegments] = segments
    const handle = await pathSegments.reduce(
      async (accum, next) => (await accum).getDirectoryHandle(next),
      this.handle.getDirectoryHandle(first)
    )
    return handle
  }

  private async getFileHandle(path: string) {
    const [segments, filename] = splitPath(path)
    const dirHandle = await this.getDirectoryHandle(segments)
    const fileHandle = await dirHandle.getFileHandle(filename)
    return fileHandle
  }

  async readTextFile(path: string) {
    const handle = await this.getFileHandle(path)
    const file = await handle.getFile()
    return file.text()
  }

  async exists(path: string): Promise<boolean> {
    if (path === "/") {
      return true
    }

    const [segments, filename] = splitPath(path)
    let { handle } = this
    for (const segment of segments) {
      try {
        handle = await handle.getDirectoryHandle(segment)
      } catch {
        return false
      }
    }

    return (
      (await success(() => handle.getFileHandle(filename))) ||
      (await success(() => handle.getDirectoryHandle(filename)))
    )
  }

  async readFile(path: string): Promise<Uint8Array> {
    const handle = await this.getFileHandle(path)
    const file = await handle.getFile()
    return new Uint8Array(await file.arrayBuffer())
  }

  async writeTextFile(path: string, data: string): Promise<void> {
    const handle = await this.getFileHandle(path)
    const writable = await handle.createWritable()
    await writable.write(data)
    await writable.close()
  }

  private async isDirectory(path: string) {
    const [segments, filename] = splitPath(path)
    const dirHandle = await this.getDirectoryHandle(segments)
    for await (const [key, handle] of dirHandle.entries()) {
      if (key !== filename) continue

      if (handle instanceof FileSystemDirectoryHandle) {
        return true
      } else {
        return false
      }
    }
    return false
  }

  async *readDir(path: string): AsyncIterable<DirEntry> {
    const segments = normalize(path).split("/")
    const dirHandle = await this.getDirectoryHandle(segments)
    for await (const key of dirHandle.keys()) {
      const isDirectory = await this.isDirectory(`${path}/${key}`)
      yield {
        name: key,
        isDirectory,
        isFile: !isDirectory,
      }
    }
  }
}

async function success(fn: () => Promise<any>) {
  try {
    await fn()
    return true
  } catch {
    return false
  }
}

/**
 * Default Browser adapter.
 */
export const getBrowserAdapter = (handle: FileSystemDirectoryHandle): Adapter => ({
  fs: new FS(handle),
  subtle: crypto.subtle,
})
