import { Buffer } from "buffer"
import type { Adapter, DirEntry, FileSystem } from "./index"

class FS implements FileSystem {
  private paths = new Set<string>()
  private pathMap = new Map<string, File>()

  constructor(list: FileList) {
    for (const file of list) {
      this.pathMap.set(file.webkitRelativePath, file)
      this.paths.add(file.webkitRelativePath)
    }
  }

  async readTextFile(path: string) {
    return this.pathMap.get(path)!.text()
  }

  async exists(path: string): Promise<boolean> {
    return this.pathMap.has(path)
  }

  async readFile(path: string): Promise<Buffer> {
    const arrayBuffer = await this.pathMap.get(path)!.arrayBuffer()
    return Buffer.from(arrayBuffer)
  }

  // eslint-disable-next-line class-methods-use-this
  async writeTextFile(): Promise<void> {
    throw new Error("fs.writeFile is not supported with webkitdirectory")
  }

  private isDirectory(path: string) {
    const paths = [...this.paths]
    return paths.some(_ => _.startsWith(`${path}/`)) && !paths.includes(path)
  }

  async *readDir(path: string): AsyncIterable<DirEntry> {
    for (const name of [...this.paths]
      .filter(_ => _.startsWith(`${path}/`))
      .map(_ => _.slice(path.length + 1))
      .map(_ => _.split("/")[0])) {
      const isDirectory = this.isDirectory(path)
      yield {
        name,
        isDirectory,
        isFile: !isDirectory,
      }
    }
  }
}

/**
 * Default Browser adapter.
 */
export const getBrowserAdapter = (list: FileList): Adapter => ({
  fs: new FS(list),
  subtle: crypto.subtle,
})
