/**
 * The buffer module from node.js, for the browser.
 *
 * @author   Feross Aboukhadijeh <https://feross.org>
 * @license  MIT
 */
// The MIT License (MIT)
// Copyright (c) 2014 Jameson Little

const lookup: string[] = []
const revLookup: number[] = []

const alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
for (let i = 0, len = alphabet.length; i < len; ++i) {
  lookup[i] = alphabet[i]
  revLookup[alphabet.charCodeAt(i)] = i
}

// Support decoding URL-safe base64 strings, as Node.js does.
// See: https://en.wikipedia.org/wiki/Base64#URL_applications
revLookup[45] = 62
revLookup[95] = 63

function base64ToByteArray(b64: string): Uint8Array {
  const { length } = b64

  if (length % 4 > 0) {
    throw new Error("Invalid string. Length must be a multiple of 4")
  }

  // Trim off extra bytes after placeholder bytes are found
  // See: https://github.com/beatgammit/base64-js/issues/42
  let validLen = b64.indexOf("=")
  if (validLen === -1) validLen = length

  const placeHoldersLen = validLen === length ? 0 : 4 - (validLen % 4)

  const arr = new Uint8Array(((validLen + placeHoldersLen) * 3) / 4 - placeHoldersLen)

  let curByte = 0

  // if there are placeholders, only get up to the last complete 4 chars
  const len = placeHoldersLen > 0 ? validLen - 4 : validLen

  let i: number
  for (i = 0; i < len; i += 4) {
    const tmp =
      (revLookup[b64.charCodeAt(i)] << 18) |
      (revLookup[b64.charCodeAt(i + 1)] << 12) |
      (revLookup[b64.charCodeAt(i + 2)] << 6) |
      revLookup[b64.charCodeAt(i + 3)]
    arr[curByte++] = (tmp >> 16) & 0xff
    arr[curByte++] = (tmp >> 8) & 0xff
    arr[curByte++] = tmp & 0xff
  }

  if (placeHoldersLen === 2) {
    const tmp =
      (revLookup[b64.charCodeAt(i)] << 2) | (revLookup[b64.charCodeAt(i + 1)] >> 4)
    arr[curByte++] = tmp & 0xff
  }

  if (placeHoldersLen === 1) {
    const tmp =
      (revLookup[b64.charCodeAt(i)] << 10) |
      (revLookup[b64.charCodeAt(i + 1)] << 4) |
      (revLookup[b64.charCodeAt(i + 2)] >> 2)
    arr[curByte++] = (tmp >> 8) & 0xff
    arr[curByte++] = tmp & 0xff
  }

  return arr
}

const tripletToBase64 = (num: number) =>
  lookup[(num >> 18) & 0x3f] +
  lookup[(num >> 12) & 0x3f] +
  lookup[(num >> 6) & 0x3f] +
  lookup[num & 0x3f]

function encodeChunk(uint8: Uint8Array, start: number, end: number) {
  const output: string[] = []
  for (let i = start; i < end; i += 3) {
    const tmp =
      ((uint8[i] << 16) & 0xff0000) +
      ((uint8[i + 1] << 8) & 0xff00) +
      (uint8[i + 2] & 0xff)
    output.push(tripletToBase64(tmp))
  }
  return output.join("")
}

export function base64FromByteArray(uint8: Uint8Array): string {
  let tmp: number
  const len = uint8.length
  const extraBytes = len % 3 // if we have 1 byte left, pad 2 bytes
  const parts = []
  const maxChunkLength = 16383 // must be multiple of 3

  // go through the array every three bytes, we'll deal with trailing stuff later
  for (let i = 0, len2 = len - extraBytes; i < len2; i += maxChunkLength) {
    parts.push(
      encodeChunk(uint8, i, i + maxChunkLength > len2 ? len2 : i + maxChunkLength)
    )
  }

  // pad the end with zeros, but make sure to not forget the extra bytes
  if (extraBytes === 1) {
    tmp = uint8[len - 1]
    parts.push(lookup[tmp >> 2] + lookup[(tmp << 4) & 0x3f] + "==")
  } else if (extraBytes === 2) {
    tmp = (uint8[len - 2] << 8) + uint8[len - 1]
    parts.push(
      lookup[tmp >> 10] + lookup[(tmp >> 4) & 0x3f] + lookup[(tmp << 2) & 0x3f] + "="
    )
  }

  return parts.join("")
}

export function fromBase64(string: string) {
  const { length } = base64ToBytes(string)
  let buf = new Uint8Array(length)

  const actual = write(buf, string)

  if (actual !== length) {
    // Writing a hex string, for example, that contains invalid characters will
    // cause everything after the first invalid character to be ignored. (e.g.
    // 'abxxcd' will be treated as 'ab')
    buf = buf.slice(0, actual)
  }

  return buf
}

function write(array: Uint8Array, string: string) {
  const offset = 0
  let { length } = array
  const remaining = array.length - offset
  if (length === undefined || length > remaining) length = remaining

  if ((string.length > 0 && (length < 0 || offset < 0)) || offset > array.length) {
    throw new RangeError("Attempt to write outside buffer bounds")
  }

  return blitBuffer(base64ToBytes(string), array, offset, length)
}

export function utf8Slice(array: Uint8Array, start = 0, end = array.length) {
  end = Math.min(array.length, end)
  const res: number[] = []

  let i = start
  while (i < end) {
    const firstByte = array[i]
    let codePoint = null
    let bytesPerSequence =
      firstByte > 0xef ? 4 : firstByte > 0xdf ? 3 : firstByte > 0xbf ? 2 : 1

    if (i + bytesPerSequence <= end) {
      let secondByte: number
      let thirdByte: number
      let fourthByte: number
      let tempCodePoint: number

      switch (bytesPerSequence) {
        case 1:
          if (firstByte < 0x80) {
            codePoint = firstByte
          }
          break
        case 2:
          secondByte = array[i + 1]
          if ((secondByte & 0xc0) === 0x80) {
            tempCodePoint = ((firstByte & 0x1f) << 0x6) | (secondByte & 0x3f)
            if (tempCodePoint > 0x7f) {
              codePoint = tempCodePoint
            }
          }
          break
        case 3:
          secondByte = array[i + 1]
          thirdByte = array[i + 2]
          if ((secondByte & 0xc0) === 0x80 && (thirdByte & 0xc0) === 0x80) {
            tempCodePoint =
              ((firstByte & 0xf) << 0xc) |
              ((secondByte & 0x3f) << 0x6) |
              (thirdByte & 0x3f)
            if (
              tempCodePoint > 0x7ff &&
              (tempCodePoint < 0xd800 || tempCodePoint > 0xdfff)
            ) {
              codePoint = tempCodePoint
            }
          }
          break
        case 4:
          secondByte = array[i + 1]
          thirdByte = array[i + 2]
          fourthByte = array[i + 3]
          if (
            (secondByte & 0xc0) === 0x80 &&
            (thirdByte & 0xc0) === 0x80 &&
            (fourthByte & 0xc0) === 0x80
          ) {
            tempCodePoint =
              ((firstByte & 0xf) << 0x12) |
              ((secondByte & 0x3f) << 0xc) |
              ((thirdByte & 0x3f) << 0x6) |
              (fourthByte & 0x3f)
            if (tempCodePoint > 0xffff && tempCodePoint < 0x110000) {
              codePoint = tempCodePoint
            }
          }
      }
    }

    if (codePoint === null) {
      // we did not generate a valid codePoint so insert a
      // replacement char (U+FFFD) and advance only 1 byte
      codePoint = 0xfffd
      bytesPerSequence = 1
    } else if (codePoint > 0xffff) {
      // encode to utf16 (surrogate pair dance)
      codePoint -= 0x10000
      res.push(((codePoint >>> 10) & 0x3ff) | 0xd800)
      codePoint = 0xdc00 | (codePoint & 0x3ff)
    }

    res.push(codePoint)
    i += bytesPerSequence
  }

  return decodeCodePointsArray(res)
}

// Based on http://stackoverflow.com/a/22747272/680742, the browser with
// the lowest limit is Chrome, with 0x10000 args.
// We go 1 magnitude less, for safety
const MAX_ARGUMENTS_LENGTH = 0x1000

function decodeCodePointsArray(codePoints: number[]) {
  const len = codePoints.length
  if (len <= MAX_ARGUMENTS_LENGTH) {
    return String.fromCharCode(...codePoints) // avoid extra slice()
  }

  // Decode in chunks to avoid "call stack size exceeded".
  let res = ""
  let i = 0
  while (i < len) {
    res += String.fromCharCode(...codePoints.slice(i, (i += MAX_ARGUMENTS_LENGTH)))
  }
  return res
}

/*
 * Need to make sure that buffer isn't trying to write out of bounds.
 */
function checkOffset(offset: number, ext: number, length: number) {
  if (offset % 1 !== 0 || offset < 0) throw new RangeError("offset is not uint")
  if (offset + ext > length) throw new RangeError("Trying to access beyond buffer length")
}

export function readUInt32BE(array: Uint8Array, offset: number, noAssert?: boolean) {
  offset = offset >>> 0
  if (!noAssert) checkOffset(offset, 4, array.length)

  return (
    array[offset] * 0x1000000 +
    ((array[offset + 1] << 16) | (array[offset + 2] << 8) | array[offset + 3])
  )
}

export function readIntLE(
  array: Uint8Array,
  offset: number,
  byteLength: number,
  noAssert?: boolean
) {
  offset = offset >>> 0
  byteLength = byteLength >>> 0
  if (!noAssert) checkOffset(offset, byteLength, array.length)

  let val = array[offset]
  let mul = 1
  let i = 0
  while (++i < byteLength && (mul *= 0x100)) {
    val += array[offset + i] * mul
  }
  mul *= 0x80

  if (val >= mul) val -= Math.pow(2, 8 * byteLength)

  return val
}

function checkInt(
  buf: Uint8Array,
  value: number,
  offset: number,
  ext: number,
  max: number,
  min: number
) {
  if (value > max || value < min) {
    throw new RangeError('"value" argument is out of bounds')
  }
  if (offset + ext > buf.length) {
    throw new RangeError("Index out of range")
  }
}

export function writeUInt32BE(
  array: Uint8Array,
  value: number,
  offset: number,
  noAssert?: boolean
) {
  value = +value
  offset = offset >>> 0
  if (!noAssert) checkInt(array, value, offset, 4, 0xffffffff, 0)
  array[offset] = value >>> 24
  array[offset + 1] = value >>> 16
  array[offset + 2] = value >>> 8
  array[offset + 3] = value & 0xff
  return offset + 4
}

// HELPER FUNCTIONS
// ================

const INVALID_BASE64_RE = /[^+/0-9A-Za-z-_]/g

function base64ToBytes(str: string) {
  // Node takes equal signs as end of the Base64 encoding
  ;[str] = str.split("=")
  // Node strips out invalid characters like \n and \t from the string, base64-js does not
  str = str.trim().replace(INVALID_BASE64_RE, "")
  // Node converts strings with length < 2 to ''
  if (str.length < 2) return new Uint8Array()
  // Node allows for non-padded base64 strings (missing trailing ===), base64-js does not
  while (str.length % 4 !== 0) {
    str = str + "="
  }
  return base64ToByteArray(str)
}

function blitBuffer(
  src: Uint8Array | number[],
  dst: Uint8Array,
  offset: number,
  length: number
) {
  let i: number
  for (i = 0; i < length; ++i) {
    if (i + offset >= dst.length || i >= src.length) break
    dst[i + offset] = src[i]
  }
  return i
}
