import { decryptData } from "./decipher"
import type { Adapter } from "./adapter"
import { createEventEmitter } from "./ee"
import { HMACAssertionError, OPVaultError } from "./errors"
import type { ItemDetails, Overview, Profile } from "./types"
import { setIfAbsent } from "./util"
import type { EncryptedItem } from "./models/item"
import { fromBase64, utf8Slice } from "./buffer"

/** Encryption and MAC */
export interface Cipher {
  /** Encryption key */
  key: Uint8Array
  /** HMAC key */
  hmac: Uint8Array
}

export class Crypto {
  #disposables: (() => void)[] = []
  #locked = true

  #master!: Cipher
  #overview!: Cipher

  private subtle: SubtleCrypto

  readonly onLock = createEventEmitter<void>()

  constructor(adapter: Adapter) {
    this.subtle = adapter.subtle
  }

  async unlock(profile: Profile, masterPassword: string) {
    const encoder = new TextEncoder()
    const key = await this.subtle.importKey(
      "raw",
      encoder.encode(masterPassword),
      { name: "PBKDF2" },
      false,
      ["deriveBits"]
    )
    const derivedKey = await this.subtle.deriveBits(
      {
        name: "PBKDF2",
        salt: fromBase64(profile.salt),
        iterations: profile.iterations,
        hash: {
          name: "SHA-512",
        },
      },
      key,
      64 << 3
    )

    const cipher = splitPlainText(new Uint8Array(derivedKey))

    // Derive master key and overview keys
    this.#master = await this.decryptKeys(profile.masterKey, cipher)
    this.#overview = await this.decryptKeys(profile.overviewKey, cipher)
    this.#locked = false
  }

  get locked() {
    return this.#locked
  }

  lock() {
    this.#locked = true
    this.#master = null!
    this.#overview = null!
    this.#disposables.forEach(fn => fn())
    this.onLock()
  }

  assertUnlocked() {
    if (this.#locked) {
      throw new OPVaultError("This vault is locked", "VAULT_LOCKED")
    }
  }

  #createCache = <K, V, K2 = K>(
    deriveArg: (value: K) => K2,
    implementation: (value: K2) => V
  ) => {
    const map = new Map<K2, V>()
    this.#disposables.push(() => map.clear())
    return (data: K) => setIfAbsent(map, deriveArg(data), implementation)
  }

  #createWeakCache = <K extends object, V>(implementation: (value: K) => V) => {
    let map = new WeakMap<K, V>()
    this.#disposables.push(() => {
      map = new WeakMap()
    })
    return (data: K) => setIfAbsent(map, data, implementation)
  }

  decryptItemDetails = this.#createWeakCache(async (item: EncryptedItem) => {
    const cipher = await this.deriveConcreteKey(item)
    const detail = await this.decryptOPData(fromBase64(item.d), cipher)
    return JSON.parse(utf8Slice(detail)) as ItemDetails
  })

  decryptItemOverview = this.#createCache(
    (item: EncryptedItem) => item.o,
    async (o: string) => {
      const overview = await this.decryptOPData(fromBase64(o), this.#overview)
      return JSON.parse(utf8Slice(overview)) as Overview
    }
  )

  deriveConcreteKey = this.#createCache(
    (data: { k: string }) => data.k,
    async ($k: string) => {
      const k = fromBase64($k)
      const data = k.slice(0, -32)
      await this.assertHMac(data, this.#master.hmac, k.slice(-32))
      const derivedKey = await this.decryptData(
        this.#master.key,
        data.slice(0, 16),
        data.slice(16)
      )
      return splitPlainText(derivedKey)
    }
  )

  async assertHMac(data: Uint8Array, key: Uint8Array, expected: Uint8Array) {
    const cryptoKey = await this.subtle.importKey(
      "raw",
      key,
      {
        name: "HMAC",
        hash: {
          name: "SHA-256",
        },
      },
      false,
      ["verify"]
    )
    const verified = await this.subtle.verify("HMAC", cryptoKey, expected, data)
    if (!verified) {
      throw new HMACAssertionError()
    }
  }

  async decryptOPData(cipherText: Uint8Array, cipher: Cipher) {
    const key = cipherText.slice(0, -32)
    await this.assertHMac(key, cipher.hmac, cipherText.slice(-32))

    const plaintext = await this.decryptData(cipher.key, key.slice(16, 32), key.slice(32))
    const size = readUint16(key.slice(8, 16))
    return plaintext.slice(-size)
  }

  async decryptData(key: Uint8Array, iv: Uint8Array, data: Uint8Array) {
    // try {
    //   const algorithm = { name: "AES-CBC", length: 256, iv }
    //   const keyCrypto = await this.subtle.importKey("raw", key, algorithm, false, [
    //     "decrypt",
    //   ])
    //   console.log("hi", keyCrypto)
    //   const decrypted = await this.subtle.decrypt(algorithm, keyCrypto, data)
    //   console.log("decrypted")
    //   return Buffer.from(decrypted)
    //   // return createDecipheriv("aes-256-cbc", key, iv).setAutoPadding(false).update(data)
    // } catch (e) {
    //   console.error(e)
    return decryptData(key, iv, data)
    // }
  }

  async decryptKeys(encryptedKey: string, derived: Cipher) {
    const buffer = fromBase64(encryptedKey)
    const base = await this.decryptOPData(buffer, derived)
    const digest = await this.subtle.digest("SHA-512", base)
    return splitPlainText(new Uint8Array(digest))
  }

  get overview() {
    return this.#overview
  }
}

export const splitPlainText = (derivedKey: Uint8Array): Cipher => ({
  key: derivedKey.slice(0, 32),
  hmac: derivedKey.slice(32, 64),
})

function readUint16({ buffer, byteOffset, length }: Uint8Array) {
  return new DataView(buffer, byteOffset, length).getUint16(0, true)
}
