import { resolve, extname, basename } from "path"
import invariant from "tiny-invariant"
import type { FileSystem } from "./adapter"
import { once } from "./util"

export type OnePasswordFileManager = ReturnType<typeof OnePasswordFileManager>

export async function OnePasswordFileManager(
  fs: FileSystem,
  path: string,
  profileName: string
) {
  const root = resolve(path, profileName)
  invariant(await fs.exists(path), `Path ${path} does not exist.`)
  invariant(await fs.exists(root), `Profile ${profileName} does not exist.`)

  const abs = (path: string) => resolve(root, path)

  const result = {
    getProfile() {
      return fs.readTextFile(abs("profile.js"))
    },

    getFolders() {
      return fs.readTextFile(abs("folders.js"))
    },

    async *getAttachments() {
      for await (const { name } of fs.readDir(root)) {
        if (extname(name) !== ".attachment") continue

        const sep = name.indexOf("_")
        const path = resolve(root, name)
        const [itemUUID, fileUUID] = [
          name.slice(0, sep),
          basename(name.slice(sep + 1), extname(name)),
        ]
        yield {
          itemUUID,
          fileUUID,
          getFile: once(() => fs.readFile(path)),
        }
      }
    },

    async getBand(name: string) {
      const path = abs(`band_${name}.js`)
      if (await fs.exists(path)) {
        return await fs.readTextFile(path)
      }
    },

    async setProfile(profile: string) {
      await fs.writeTextFile("profile.js", profile)
    },

    async setFolders(folders: string) {
      await fs.writeTextFile("folders.js", folders)
    },

    async setBand(name: string, band: string) {
      await fs.writeTextFile(`band_${name}.js`, band)
    },
  }
  return result
}
