import { resolve } from "path"
import { Vault } from "./models/vault"
import type { Adapter } from "./adapter"

export type { Vault } from "./models/vault"
export type { Item } from "./models/item"
export type { Attachment, AttachmentMetadata } from "./models/attachment"
export type { ItemField, ItemSection } from "./types"
export { Category, FieldType } from "./models"

export type { Adapter as IAdapter } from "./adapter/index"

interface IOptions {
  /**
   * Path to `.opvault` directory
   */
  path: string

  /**
   * Adapter used to interact with the file system and cryptography modules
   */
  adapter: Adapter | Promise<Adapter>
}

/**
 * OnePassword instance
 * @example
 * ```ts
 * import { OnePassword } from "opvault.js"
 * import { adapter } from "opvault.js/src/adapter/node"
 *
 * const op = new OnePassword({
 *   path: "/path/to/1password/vault",
 *   adapter,
 * })
 *
 * const profileNames = await op.getProfileNames()
 * const vault = await op.getProfile(profileNames[0])
 * const item = await vault.getItemByTitle("My Login")
 * ```
 */
export class OnePassword {
  readonly #path: string
  readonly #adapter: Adapter | Promise<Adapter>

  constructor(options: IOptions) {
    this.#adapter = options.adapter
    this.#path = options.path
  }

  /**
   * @returns A list of names of profiles of the current vault.
   */
  async getProfileNames() {
    const { fs } = await this.#adapter
    const profiles: string[] = []

    for await (const { name, isDirectory } of fs.readDir(this.#path)) {
      const fullPath = resolve(this.#path, name)
      if (isDirectory && (await fs.exists(resolve(fullPath, "profile.js")))) {
        profiles.push(name)
      }
    }

    return profiles
  }

  /**
   * @returns A OnePassword Vault instance.
   */
  async getProfile(profileName: string) {
    return await Vault.of(this.#path, profileName, await this.#adapter)
  }
}
