import type { Crypto } from "../crypto"
import { invariant } from "../errors"
import { fromBase64, readIntLE, utf8Slice } from "../buffer"

type integer = number

export interface AttachmentMetadata {
  itemUUID: string
  contentSize: integer
  external: boolean
  updatedAt: integer
  txTimestamp: integer
  /** Encoded as base64 encoded OPData */
  overview: {
    filename: string
  }
  createdAt: integer
  uuid: string
}

export class Attachment {
  #k: string
  #crypto: Crypto
  #buffer: Uint8Array

  #icon?: Uint8Array // png buffer
  #file?: Uint8Array
  #metadata?: AttachmentMetadata

  private metadataSize: number
  private iconSize: number

  constructor(crypto: Crypto, k: string, buffer: Uint8Array) {
    this.#buffer = buffer
    this.#validate()
    this.#crypto = crypto
    this.#k = k
    this.metadataSize = readIntLE(buffer, 8, 2)
    this.iconSize = readIntLE(buffer, 12, 3)

    crypto.onLock(() => {
      this.#lock()
    })
  }

  /**
   * Validate attachment file.
   */
  #validate() {
    const file = this.#buffer
    invariant(
      utf8Slice(file.slice(0, 6)) === "OPCLDA",
      "Attachment must start with OPCLDA"
    )
    // @TODO: Re-enable this
    false &&
      invariant(
        readIntLE(file, 7, 1) === 1,
        "The version for this attachment file format is not supported."
      )
  }

  get icon() {
    return this.#icon!
  }

  get file() {
    return this.#file!
  }

  get metadata() {
    return this.#metadata!
  }

  async unlock() {
    const crypto = this.#crypto
    const cipher = await crypto.deriveConcreteKey({ k: this.#k })
    const { metadataSize, iconSize } = this
    const buffer = this.#buffer
    this.#icon = await crypto.decryptOPData(
      buffer.slice(16 + metadataSize, 16 + metadataSize + iconSize),
      cipher
    )
    this.#file = await crypto.decryptOPData(
      buffer.slice(16 + metadataSize + iconSize),
      cipher
    )

    const metadata = JSON.parse(utf8Slice(buffer.slice(16, 16 + metadataSize)))
    metadata.overview = JSON.parse(
      utf8Slice(
        await crypto.decryptOPData(fromBase64(metadata.overview), crypto.overview)
      )
    )
    this.#metadata = metadata
  }

  #lock() {
    this.#metadata = undefined!
    this.#icon = undefined!
    this.#file = undefined!
  }

  dispose() {
    this.#buffer = null!
    this.#lock()
  }
}
