import styled from "@emotion/styled"
import { useEffect, useMemo, useState } from "react"
import type { Item } from "opvault.js"
import { Category } from "opvault.js"
import { IoSearch } from "react-icons/io5"
import { ItemList } from "../components/ItemList"
import { ItemView } from "../components/Item"
import { reactIconClass } from "../components/CategoryIcon"
import { useTranslate } from "../i18n/index"
import { scrollbar } from "../styles"

const ListContainer = styled.div`
  border-right: 1px solid var(--border-color);
  width: 350px;
  margin-right: 10px;
  overflow-y: scroll;
  overflow-y: overlay;
  overflow-x: hidden;
  @media (prefers-color-scheme: dark) {
    background: #202020;
    border-right-color: transparent;
  }
`
const ItemContainer = styled.div`
  width: calc(100% - 300px);
  overflow: hidden;
`
const SearchContainer = styled.div`
  text-align: center;
  position: relative;
  flex-grow: 1;
  margin: 10px 0;
  margin-right: 10px;
`
const SortContainer = styled.div`
  display: flex;
  margin: 10px 0;
`
const CategorySelect = styled.select`
  width: 50%;
  margin-left: 10px;
  margin-right: 5px;
`
const SortSelect = styled.select`
  width: calc(50% - 25px);
`

const SearchInput = styled.input`
  --margin: 10px;
  width: calc(100% - var(--margin) * 2 + 9px);
  margin: 0 var(--margin);
  padding-left: 2em !important;
`
const SearchIcon = styled(IoSearch)`
  position: absolute;
  top: 9px;
  left: 20px;
`

const enum SortBy {
  Name,
  CreatedAt,
  UpdatedAt,
}

export const FilteredVaultView: React.FC<{ items: Item[] }> = ({ items }) => {
  const t = useTranslate()
  const [item, setItem] = useState<Item>()
  const [category, setCategory] = useState<Category>()
  const [sortBy, setSortBy] = useState(SortBy.Name)
  const [search, setSearch] = useState("")

  useEffect(() => {
    setItem(undefined)
  }, [items])

  const compareFn = useMemo((): ((a: Item, b: Item) => number) => {
    switch (sortBy) {
      case SortBy.Name:
        return (a, b) => (a.overview.title ?? "").localeCompare(b?.overview.title ?? "")
      case SortBy.CreatedAt:
        return (a, b) => b.createdAt - a.createdAt
      case SortBy.UpdatedAt:
        return (a, b) => b.updatedAt - a.updatedAt
    }
  }, [sortBy])

  const sortedItem = useMemo(() => items.slice().sort(compareFn), [items, compareFn])

  const filtered = useMemo(() => {
    let items = sortedItem.filter(x => x.category !== Category.Tombstone)
    if (category != null) {
      items = items.filter(x => x.category === category)
    }

    let res: Item[] = items
    if (search) {
      res = []
      for (const x of items) {
        const compare = Math.max(
          stringCompare(search, x.overview.title),
          stringCompare(search, x.overview.ainfo)
        ) as CompareResult
        switch (compare) {
          case CompareResult.NoMatch:
            continue
          case CompareResult.Includes:
            res.push(x)
            break
          case CompareResult.Equals:
            res.unshift(x)
            break
        }
      }
    }
    return res
  }, [sortedItem, search, category])

  const categoryMap = useMemo(
    (): [Category | undefined, string][] => [
      [undefined, t.label.category_all],
      [Category.Login, t.label.category_login],
      [Category.SecureNote, t.label.category_secure_note],
      [Category.CreditCard, t.label.category_credit_card],
      [Category.Identity, t.label.category_identity],
      [Category.Password, t.label.category_password],
      [Category.Membership, t.label.category_membership],
      [Category.Database, t.label.category_database],
      [Category.BankAccount, t.label.category_bank_account],
      [Category.Email, t.label.category_email],
      [Category.SoftwareLicense, t.label.category_software_license],
      [Category.SSN, t.label.category_ssn],
      [Category.Passport, t.label.category_passport],
      [Category.OutdoorLicense, t.label.category_outdoor_license],
      [Category.DriverLicense, t.label.category_driver_license],
      [Category.Rewards, t.label.category_rewards],
      [Category.Router, t.label.category_router],
      [Category.Server, t.label.category_server],
    ],
    [t]
  )

  return (
    <>
      <ListContainer className={scrollbar}>
        <SearchContainer>
          <SearchInput
            type="search"
            value={search}
            onChange={e => setSearch(e.currentTarget.value)}
          />
          <SearchIcon className={reactIconClass} />
        </SearchContainer>

        <SortContainer>
          <CategorySelect
            value={category}
            onChange={e => setCategory((e.currentTarget.value as Category) || undefined)}
          >
            {categoryMap.map(([value, name]) => (
              <option value={value || ""} key={name}>
                {name}
              </option>
            ))}
          </CategorySelect>

          <SortSelect value={sortBy} onChange={e => setSortBy(+e.currentTarget.value)}>
            <option key={1} value={SortBy.Name}>
              {t.options.sort_by_name}
            </option>
            <option key={2} value={SortBy.CreatedAt}>
              {t.options.sort_by_created_at}
            </option>
            <option key={3} value={SortBy.UpdatedAt}>
              {t.options.sort_by_updated_at}
            </option>
          </SortSelect>
        </SortContainer>
        <ItemList items={filtered} onSelect={setItem} selected={item} />
      </ListContainer>
      <ItemContainer>
        {item && <ItemView className={scrollbar} item={item} />}
      </ItemContainer>
    </>
  )
}

enum CompareResult {
  NoMatch,
  Includes,
  Equals,
}

function stringCompare(search: string, source?: string) {
  if (!search) return CompareResult.Includes
  if (!source) return CompareResult.NoMatch
  source = source.toLocaleLowerCase()
  search = search.toLocaleUpperCase()
  const includes = source.includes(search.toLocaleLowerCase())
  if (includes) {
    return source.length === search.length ? CompareResult.Equals : CompareResult.Includes
  }
  return CompareResult.NoMatch
}
