import styled from "@emotion/styled"
import { css } from "@emotion/css"
import { useCallback, useMemo, memo, useState } from "react"
import { Si1Password } from "react-icons/si"
import { FaFolderOpen } from "react-icons/fa"
import { ImCross } from "react-icons/im"
import { MdClearAll } from "react-icons/md"
import { BsGear, BsInfoCircle } from "react-icons/bs"
import { openDirectory } from "../../utils/ipc-adapter"
import { useTranslate } from "../../i18n"
import { Key, useStorage } from "../../utils/localStorage"
import { Settings } from "../../settings"
import { About } from "../../about"

const Container = styled.div`
  padding: 100px;
  max-width: 600px;
  margin: 0 auto;
`
const List = styled.ul`
  list-style-type: none;
  padding-inline-start: 0;
`
const Item = styled.li`
  align-items: center;
  cursor: default;
  display: flex;
  padding: 8px 10px;
  user-select: none;
  &:not(:hover):not(:active) {
    background: transparent;
    border-color: transparent;
  }
`
const icon = css`
  font-size: 1.5em;
  margin-right: 10px;
`
const Text = styled.div`
  flex-grow: 1;
`
const Hr = styled.hr`
  border: none;
  border-top: 1px solid var(--border-color);
`
const DeleteItem = styled(ImCross)`
  text-align: right;
  font-size: 0.7em;
  opacity: 0;
  ${Item}:hover & {
    opacity: 1;
  }
`
const NonCriticalPath = styled.span`
  opacity: 0.4;
`

const Path = memo(({ children }: { children: string }) => {
  const segments = useMemo(() => children.split("/"), [children])
  return (
    <span>
      {segments.map((seg, i, { length }) =>
        i < length - 1 ? (
          <NonCriticalPath key={i}>{seg}/</NonCriticalPath>
        ) : (
          <span key={i}>{seg}</span>
        )
      )}
    </span>
  )
})

const enum Modal {
  None,
  Settings,
  About,
}

export const PickOPVault: React.FC<{
  setPath(path: string): void
}> = ({ setPath }) => {
  const t = useTranslate()
  const [modal, setModal] = useState(Modal.None)
  const [list, $setList] = useStorage(Key.RECENTLY_OPENED_VAULTS)

  const clearHistory = useCallback(() => {
    $setList([])
  }, [$setList])

  const setList = useCallback(
    (fn: (value: Set<string>) => void) => {
      $setList(list => {
        const set = new Set(list)
        fn(set)
        return Array.from(set)
      })
    },
    [$setList]
  )

  const onClick = useCallback(async () => {
    const path = await openDirectory()
    if (path) {
      setPath(path)
      setList(set => set.add(path))
    }
  }, [setPath, setList])

  return (
    <Container>
      <List>
        <Item className="button" onClick={onClick}>
          <FaFolderOpen className={icon} />
          {t.label.choose_a_vault}…
        </Item>
        {list.map((item, i) => (
          <Item className="button" onClick={() => setPath(item)} key={i}>
            <Si1Password className={icon} />
            <Text>
              <Path>{item}</Path>
            </Text>
            <DeleteItem onClick={() => setList(list => list.delete(item))} />
          </Item>
        ))}
        {list.length > 0 && (
          <>
            <Hr />
            <Item className="button" onClick={() => clearHistory()}>
              <MdClearAll className={icon} />
              {t.action.clear_history}
            </Item>
          </>
        )}
        <Item className="button" onClick={() => setModal(Modal.Settings)}>
          <BsGear className={icon} />
          {t.label.settings}
        </Item>
        <Item className="button" onClick={() => setModal(Modal.About)}>
          <BsInfoCircle className={icon} />
          {t.label.about_app}
        </Item>
      </List>

      <Settings show={modal === Modal.Settings} onHide={() => setModal(Modal.None)} />
      <About show={modal === Modal.About} onHide={() => setModal(Modal.None)} />
    </Container>
  )
}
